/****************************************************************************
 *
 *           Program : pi-mc.c
 *
 *	     Author  : Kim, Dongmin
 *
 *
 *  This is a program that calculate pi by using Monte Carlo technique.
 *
 *  The pi is as the are of a circle of radius 1 (in two dimensions).
 *
 *  We can set this up to use one quadrant of the circle and then multiply
 *
 *  the result by 4. Thus pi should be 4 * (mean of f), where f is 
 *
 *       f(x, y) = 1, if ( x**2 + y**2) <= 1
 *               = 0, otherwise
 *
 *  In this program, each processor generates pi independently and
 *
 *  0 processor collects them and calculates the derivation.
 *
 *  For derivation, I used the formula
 *
 *  derivation = sqrt [ { sigma ( x^2) } / N - average ^ 2 ]
 *
 *  To complie : gcc pi-mc.c -o pi -L/usr/local/lib/rs6000/ch_p4 -lmpi -lm
 *
 *  To run     : mpirun -mr_machinefile machinelist -mr_np 4 pi
 *
 ****************************************************************************/

#include	<stdio.h>
#include	<mpi.h>
#include        <math.h>

#define MAX     10000000
#define G_start 0.0  /* global start point */
#define G_end   1.0  /* global end point */

#define real_pi   ( (double) (4.0*atan(1.0)) ) /* the calculate value of PI */

/****************************************************************************
 *
 *         Procedure : f()
 *         Input     : double x double y
 *         Output    : double 
 *         Function  : calculate x**2 + y**2 and return this result.
 *
 ****************************************************************************/
double  f( x, y )
     double x, y;
{
  return ( x*x + y*y );
}

/****************************************************************************
 *
 *         Procedure : pi()
 *         Input     : double start_pt, double end_pt
 *         Output    : double summation
 *         Function  : find the summation between start_pt and end_pt
 *                     using Monte Carlo technique.
 *                     In this program, MAX numbers are generated in each
 *                     processor.
 *
 ****************************************************************************/
double pi(seed)
     long   seed;
{
  int    i;
  double x, y;
  double sum;
  double temp;

  sum = 0.;
  srand48(seed);

  for ( i = 0 ; i < MAX ; i++ ) {
    x = (double) drand48();
    y = (double) drand48();

    temp = f(x, y);
    if ( temp <= 1 ) {
      sum = sum + 1.;
    }
  }
  return sum;

}

/****************************************************************************
 *
 *         Procedure : main()
 *         Function  : main routine of this program.
 *
 ****************************************************************************/
void main(argc, argv)
  int argc;
  char **argv;
{

   int    nprocs;      /* Number of processors */
   int    pnum;        /* Processor number starting from 0 */
   long   seed;        /* seed for random number */
   double local_start; /* local start point */
   double local_end;   /* local end point */
   double term;        /* the length of term */
   double ave;         /* average between local_start and local_end */
   double other;       /* result from other processor */
   double error;       /* difference */
   int    from, i;
   double sum_2, sum, var, dev;
   int    sequence;

   MPI_Status  recv_status;

   MPI_Init(&argc, &argv);

   /* Who am I ? */
   MPI_Comm_rank(MPI_COMM_WORLD, &pnum);

   /* How many processors participate ? */
   MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

   sum_2 = 0.;
   sum = 0.;

   /* setup the seed using processor number */
   seed = (long) (pnum);

   /* calculate the sum */
   ave = 4. * pi( seed ) / MAX;

   printf("Processor : %d      pi : %.18f\n", pnum, ave);

   if ( pnum == 0 ) { /* 0 processor collects the sums from other processors */

     sum_2 = ave * ave;
     ave = ave;

     for ( from = 1 ; from < nprocs ; from++ ) {

       /* 0 processor receives the sums from the other processors */
       MPI_Recv( &other, 1, MPI_DOUBLE, from, 99, MPI_COMM_WORLD, &recv_status);

       ave = ave + other;

       /* calculate the sigma of x ^ 2 */
       sum_2 = sum_2 + other * other;
     }

     ave = ave / (double) nprocs;

     var = ( sum_2 / (double) nprocs ) - ave * ave;

     dev = sqrt(var);

     error = real_pi - ave;
     printf("\n\n\n");
     printf("***************************************************************\n");
     printf("  PI using Monte Carlo Technique \n");
     printf("\n");
     printf("   Number of Processor(s) : %d\n", nprocs);
     printf("   PI ( by Monte Carlo ) = %.18f\n", ave);
     printf("   PI ( by calculation ) = %.18f\n", real_pi);
     printf("   Variance   : %.18f\n", var );
     printf("   Derivation : %.18f\n", dev );
     printf("   error = %18.16f\n", error);
     printf("***************************************************************\n");

   }
   else { /* Processors which is not 0 send the sum to the 0 processor */
     MPI_Send( &ave, 1, MPI_DOUBLE, 0, 99, MPI_COMM_WORLD);
   }

   MPI_Finalize();

}
